# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Handle the product specification
================================
"""
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union
import copy
import enum
import os
import xml.etree.ElementTree as xt

#
import numpy as np
import numpy.typing as npt
import xarray as xr

#
from . import EXPERT, PRODUCT_TYPE, UNSMOOTHED
from . import math


class Side(enum.Enum):
    """Represents both sides of the swath."""
    LEFT = "left"
    RIGHT = "right"


def _find(element: xt.Element, tag: str) -> xt.Element:
    """Find a tag in the xml format specifcation file"""
    result = element.find(tag)
    if result is None:
        raise RuntimeError("The XML tag '" + tag + "' doesn't exist")
    return result


def _parse_type(dtype, width, signed) -> Union[npt.DTypeLike, np.generic]:
    """Parse type from xml format specification file. """
    if dtype == "real":
        return getattr(np, "float" + width)
    if dtype == "integer":
        return getattr(np, ("u" if not signed else "") + "int" + width)
    if dtype == "string":
        return np.str_
    if dtype == "char":
        return np.dtype(f"S{width}")  # type: ignore
    raise ValueError("Data type '" + dtype + "' is not recognized.")


def cast_to_dtype(attr_value: Union[int, float], properties: Dict[str, str]):
    """Cast the attribute value to the numpy data type required"""
    return getattr(np, properties["dtype"])(attr_value)


def _strtobool(value: str) -> bool:
    """Return the boolean value encoded in the string"""
    value = value.lower()
    if value == "true":
        return True
    if value == "false":
        return False
    raise ValueError(f"invalid truth value {value!r}")


def _strip_shape(name: str) -> str:
    """Strip the dimension name."""
    if name[0] == "/":
        return name[1:]
    return name


def _parser(tree: xt.ElementTree):
    """Parse variables, attributes and shapes from xml format specification
    file"""
    variables = dict()
    attributes = dict()
    shapes = dict()

    for item in tree.getroot().findall("shape"):
        dims = tuple(
            _strip_shape(dim.attrib["name"])
            for dim in item.findall("dimension"))
        if dims:
            shapes[item.attrib["name"]] = dims

    for item in _find(_find(tree.getroot(), 'science'), 'nodes'):
        dtype = _parse_type(
            item.tag, item.attrib["width"],
            _strtobool(item.attrib["signed"])
            if "signed" in item.attrib else None)
        if not isinstance(dtype, np.dtype):
            dtype = dtype.__name__  # type: ignore
        annotation = item.find("annotation")
        if annotation is None:
            continue
        varname = item.attrib["name"]
        if varname.startswith("/@"):
            attributes[varname[2:]] = dict(attrs=annotation.attrib,
                                           dtype=dtype)
        else:
            del annotation.attrib["app"]
            variables[varname[1:]] = dict(attrs=annotation.attrib,
                                          dtype=dtype,
                                          shape=shapes[item.attrib["shape"]])

    return variables, attributes


def parse_specification_file(path: str) -> Tuple:
    """Parse the XML specification file"""
    return _parser(xt.parse(path))


def encode_fill_value(properties: Dict[str, Any]) -> np.ndarray:
    """Returns the fill value encoded with the data type specified"""
    dtype = properties["dtype"]
    attrs = properties["attrs"]
    fill_value = attrs["_FillValue"] if "_FillValue" in attrs else 0
    if isinstance(dtype, str):
        return getattr(np, dtype)(fill_value)
    return np.array(fill_value, dtype)


def build_array(name: str, variables: Dict[str, Dict[str, Any]],
                data: np.ndarray):
    """Builds the array from the data and the variables"""
    properties = variables[name]
    attrs = copy.deepcopy(properties["attrs"])

    # Reading the storage properties of the variable
    encoding: Dict[str, Any] = dict(dtype=properties["dtype"])

    # If the variable defines a fill value.
    if "_FillValue" in attrs:
        encoding["_FillValue"] = encode_fill_value(properties)
        del attrs["_FillValue"]

    # Some values read from the XML files must be decoded
    # TODO(fbriol): The type of these attributes should be determined
    # from their type, but at the moment this is not possible.
    for item in ["add_offset", "scale_factor"]:
        if item in attrs:
            attrs[item] = float(attrs[item])
    for item in ["valid_range", "valid_min", "valid_max"]:
        if item in attrs:
            attrs[item] = cast_to_dtype(attrs[item], properties)
    if "flag_values" in attrs:
        items = attrs["flag_values"].split()
        attrs["flag_values"] = np.array(
            [cast_to_dtype(item, properties) for item in items],
            properties["dtype"]) if len(items) != 1 else cast_to_dtype(
                float(attrs["flag_values"]), properties)
    # if "scale_factor" in attrs and "add_offset" not in attrs:
    #     attrs["add_offset"] = 0.0
    # if "add_offset" in attrs and "scale_factor" not in attrs:
    #     attrs["scale_factor"] = 1.0
    return {
        name: encoding
    }, xr.DataArray(data=data,
                    dims=properties["shape"],
                    name=name,
                    attrs=attrs)


class ProductSpecification:
    """Parse and load into memory the product specification"""
    SPECIFICATION = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 "l2b-ssh.xml")

    def __init__(self, product_type: Optional[str]):
        product_type = product_type or EXPERT
        self.unsmoothed = product_type == UNSMOOTHED
        self.specification = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            PRODUCT_TYPE[product_type])
        self.variables, self.attributes = parse_specification_file(
            self.specification)

    def __contains__(self, item):
        if self.unsmoothed:
            item = f"{Side.LEFT.value}/{item}"
        return item in self.variables

    def ssh_karin_name(self) -> str:
        """Get the name of the variable containing the SSH."""
        return "ssh_karin" + ("_2" if self.unsmoothed else "")

    def _variable(self, name: str):
        """Get the name of the template variable."""
        return self.variables[f"{Side.LEFT.value}/{name}" if self.
                              unsmoothed else name]

    def _shape(self, swath: bool = True):
        """Get the shape of the variable"""
        if swath:
            return self._variable("longitude")["shape"]
        return self._variable("time")["shape"]

    def num_sides(self) -> Dict[str, int]:
        """Get the name of the dimensions representing the number of sides."""
        if self.unsmoothed:
            return {
                f"{Side.LEFT.value}/num_sides": 1,
                f"{Side.RIGHT.value}/num_sides": 1
            }
        return {"num_sides": 2}

    def _names(self, name: str, split: bool = True) -> Tuple[str, ...]:
        """Get the names of the variables in the dataset."""
        if self.unsmoothed and split:
            return (f"{Side.LEFT.value}/{name}", f"{Side.RIGHT.value}/{name}")
        return (name, )

    def _data_array(
            self,
            name: str,
            data: np.ndarray,
            split: bool = True) -> Optional[Tuple[Dict, List[xr.DataArray]]]:
        """Returns a tuple containing the netCDF encoding of the variable and
        the data array."""
        result = {}, []

        if split and self.unsmoothed:
            if len(data.shape) == 2:
                middle = data.shape[1] // 2
                swaths = (data[:, :middle], data[:, middle:])
            else:
                swaths = (data, data)
            names = self._names(name)
        else:
            swaths = (data, )
            names = self._names(name, split=False)

        for ix, varname in enumerate(names):
            if varname not in self.variables:
                return None
            encoding_array, array = build_array(varname, self.variables,
                                                swaths[ix])
            result[0].update(encoding_array)
            result[1].append(array)
        return result if result[1] else None

    def time(self, time: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Gets the time axis"""
        return self._data_array("time", time)  # type: ignore

    def x_ac(self,
             x_ac: np.ndarray) -> Optional[Tuple[Dict, List[xr.DataArray]]]:
        """Returns the properties of the variable describing the cross track
        distance"""
        return self._data_array("cross_track_distance", x_ac * 1000)  # km -> m

    def lon(self, lon: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the longitudes of
        the swath"""
        # Longitude must be in [0, 360.0[
        result = self._data_array("longitude",
                                  math.normalize_longitude(lon, 0))
        assert result is not None
        return result

    def lon_nadir(
            self, lon_nadir: np.ndarray
    ) -> Optional[Tuple[Dict, List[xr.DataArray]]]:
        """Returns the properties of the variable describing the longitudes of
        the reference ground track"""
        # Longitude must be in [0, 360.0[
        return self._data_array("longitude_nadir",
                                math.normalize_longitude(lon_nadir, 0))

    def lat(self, lat: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the latitudes of
        the swath"""
        result = self._data_array("latitude", lat)
        assert result is not None
        return result

    def lat_nadir(
            self, lat_nadir: np.ndarray
    ) -> Optional[Tuple[Dict, List[xr.DataArray]]]:
        """Returns the properties of the variable describing the latitudes of
        the reference ground track"""
        return self._data_array("latitude_nadir", lat_nadir)

    def ssh_karin(self, ssh: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the SSH measured
        by KaRIn"""
        result = self._data_array(self.ssh_karin_name(), ssh)
        assert result is not None
        return result

    def ssh_nadir(self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the SSH to
        nadir."""
        name = "ssh_nadir"
        for item in self._names(name, split=False):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                coordinates='longitude latitude',
                long_name='sea surface height',
                scale_factor=0.0001,
                standard_name='seaurface height above reference ellipsoid',
                units='m',
                valid_min=np.int32(-15000000),
                valid_max=np.int32(150000000)),
                                        dtype='int32',
                                        shape=self._shape(swath=False))
        return self._data_array(name, array)  # type: ignore

    def simulated_true_ssh_nadir(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the SSH to nadir
        free of measurement errors."""
        name = "simulated_true_ssh_nadir"
        for item in self._names(name, split=False):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                coordinates='time',
                long_name='sea surface height',
                scale_factor=0.0001,
                standard_name='sea surface height above reference ellipsoid',
                units='m',
                valid_min=np.int32(-15000000),
                valid_max=np.int32(150000000),
                comment='Height of the sea surface free of measurement errors.'
            ),
                                        dtype='int32',
                                        shape=self._shape(swath=False))
        return self._data_array(name, array)  # type: ignore

    def swh_karin(self, swh: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the SWH measured
        by KaRIn"""
        result = self._data_array("swh_karin", swh)
        assert result is not None
        return result

    def swh_nadir(self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the SWH to nadir
        free of measurement errors."""
        name = "swh_nadir"
        for item in self._names(name, split=False):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                coordinates='time',
                long_name='Significant Wave Height',
                scale_factor=0.0001,
                standard_name='Sigificant Wave Height',
                units='m',
                valid_min=np.int32(-15000000),
                valid_max=np.int32(150000000),
            ),
                                        dtype='int32',
                                        shape=self._shape(swath=False))
        return self._data_array(name, array)  # type: ignore

    def simulated_true_ssh_karin(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the SSH KaRIn free
        of measurement errors."""
        name = "simulated_true_ssh_karin"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                coordinates='longitude latitude',
                long_name='sea surface height',
                scale_factor=0.0001,
                standard_name='sea surface height above reference ellipsoid',
                units='m',
                valid_min=np.int32(-15000000),
                valid_max=np.int32(150000000),
                comment='Height of the sea surface free of measurement errors.'
            ),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_baseline_dilation(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the error due to
        baseline mast dilation"""
        name = "simulated_error_baseline_dilation"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Error due to baseline mast dilation',
                scale_factor=0.0001,
                units='m',
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_roll(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the error due to
        roll"""
        name = "simulated_error_roll"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Error due to roll',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_phase(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the error due to
        phase"""
        name = "simulated_error_phase"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Error due to phase',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_orbital(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the error due to
        orbital perturbations."""
        name = "simulated_error_orbital"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Error due to orbital perturbations',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_roll_phase_estimate(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the roll phase
        correction estimated"""
        name = "simulated_roll_phase_estimate"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Error after estimation of roll phase',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_karin(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the KaRIn error"""
        name = "simulated_error_karin"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='KaRIn error',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_timing(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the timing
        error"""
        name = "simulated_error_timing"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Timing error',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_troposphere(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the error due to
        wet troposphere path delay"""
        name = "simulated_error_troposphere"
        for item in self._names(name):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Error due to wet troposphere path delay',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape())
        return self._data_array(name, array)  # type: ignore

    def simulated_error_troposphere_nadir(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the error due to
        wet troposphere path delay to nadir."""
        name = "simulated_error_troposphere_nadir"
        for item in self._names(name, split=False):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Error due to wet troposphere path delay',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape(swath=False))
        return self._data_array(name, array)  # type: ignore

    def simulated_error_altimeter(
            self, array: np.ndarray) -> Tuple[Dict, List[xr.DataArray]]:
        """Returns the properties of the variable describing the altimeter
        error"""
        name = "simulated_error_altimeter"
        for item in self._names(name, split=False):
            self.variables[item] = dict(attrs=dict(
                _FillValue=2147483647,
                long_name='Altimeter error',
                standard_name='',
                units='m',
                scale_factor=0.0001,
                coordinates='longitude latitude'),
                                        dtype='int32',
                                        shape=self._shape(swath=False))
        return self._data_array(name, array)  # type: ignore

    def fill_variables(self, variables,
                       shape) -> Iterator[Tuple[Dict, List[xr.DataArray]]]:
        """Returns the properties of variables present in the official format
        of the SWOT product, but not calculated by this software."""
        for item, properties in self.variables.items():
            if item in variables:
                continue
            fill_value = encode_fill_value(properties)
            data = np.full(tuple(shape[dim] for dim in properties["shape"]),
                           fill_value, properties["dtype"])
            variable = self._data_array(item, data, split=False)
            if variable is not None:
                yield variable
        return StopIteration



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Roll errors
-----------
"""
from typing import Dict, Optional, Tuple
import numpy as np

from .. import random_signal
from .. import settings
from .. import VOLUMETRIC_MEAN_RADIUS, CELERITY, F_KA, BASELINE


class RollPhase:
    """
    Roll errors

    Args:
        parameters (settings.Parameters): Simulation settings
        roll_psd (numpy.ndarray): Power spectral density the roll control angle
        gyro_psd (numpy.ndarray): Power spectral density the roll error
            knowledge (also called gyro error)
        phase_psd (numpy.ndarray): Power spectral density the error angle
        spatial_frequency (numpy.ndarray): Spatial frequency
    """
    def __init__(self, parameters: settings.Parameters, roll_psd: np.ndarray,
                 gyro_psd: np.ndarray, phase_psd: np.ndarray,
                 spatial_frequency: np.ndarray) -> None:
        # Store the generation parameters of the random signal.
        self.rng_theta = parameters.rng()
        self.rng_theta_l = parameters.rng()
        self.rng_theta_r = parameters.rng()
        self.len_repeat = parameters.len_repeat
        self.delta_al = parameters.delta_al

        # Get baseline dilation power spectrum
        self.roll_psd = roll_psd + gyro_psd
        indexes = np.where(spatial_frequency <= 1 / 40000)[0]
        self.roll_psd[indexes] = self.roll_psd[indexes][-1]
        self.phase_psd = phase_psd
        self.spatial_frequency = spatial_frequency

        # TODO
        assert parameters.height is not None
        height = parameters.height * 1e-3
        self.phase_conversion_factor = (
            1 / (F_KA * 2 * np.pi / CELERITY * BASELINE) *
            (1 + height / VOLUMETRIC_MEAN_RADIUS) * np.pi / 180 * 1e3)
        self.roll_conversion_factor = (
            (1 + height / VOLUMETRIC_MEAN_RADIUS) * np.pi / 180 / 3600) * 1e3

    def _generate_1d(self, x_al: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:

        # Compute roll angle using the power spectrum
        # Compute left and right phase angles the power spectrum
        theta = random_signal.gen_signal_1d(self.spatial_frequency,
                                            self.roll_psd,
                                            x_al,
                                            rng=self.rng_theta,
                                            fmin=1 / self.len_repeat,
                                            fmax=1 / (2 * self.delta_al),
                                            alpha=10)
        theta_l = random_signal.gen_signal_1d(self.spatial_frequency,
                                              self.phase_psd,
                                              x_al,
                                              rng=self.rng_theta_l,
                                              fmin=1 / self.len_repeat,
                                              fmax=1 / (2 * self.delta_al),
                                              alpha=10)
        theta_r = random_signal.gen_signal_1d(self.spatial_frequency,
                                              self.phase_psd,
                                              x_al,
                                              rng=self.rng_theta_r,
                                              fmin=1 / self.len_repeat,
                                              fmax=1 / (2 * self.delta_al),
                                              alpha=10)
        # Compute the associated roll  error on the swath in m
        roll = self.roll_conversion_factor * theta
        # Compute the associated phase error on the swath in m
        phase = np.array([
            self.phase_conversion_factor * theta_l.T,
            self.phase_conversion_factor * theta_r.T
        ])
        return roll, phase.T

    def generate(
        self,
        x_al: np.ndarray,
        x_ac: np.ndarray,
    ) -> Dict[str, np.ndarray]:
        """Generate roll and phase errors

        Args:
            x_al (numpy.ndarray): Along track distance
            x_ac (numpy.ndarray): Across track distance

        Returns:
            dict: variable name and errors simulated.
        """
        roll_1d, phase_1d = self._generate_1d(x_al)
        num_pixels = x_ac.shape[0]
        swath_center = num_pixels // 2
        ac_l = x_ac[:swath_center]
        ac_r = x_ac[swath_center:]

        phase = np.full((phase_1d.shape[0], num_pixels), np.nan)
        phase[:, :swath_center] = ac_l * phase_1d[:, 0, np.newaxis]
        phase[:, swath_center:] = ac_r * phase_1d[:, 1, np.newaxis]

        # rollphase_est_1d = np.zeros((np.shape(phase_1d.T)))
        # rollphase_est = np.full((np.shape(rollphase_est_1d)[0], nac), np.nan)
        # rollphase_est[:, :mid_swath] = np.mat(rollphase_est_1d[:, 0]).T * ac_l
        # rollphase_est[:, mid_swath:] = np.mat(rollphase_est_1d[:, 1]).T * ac_r
        return {
            "simulated_error_roll": x_ac * roll_1d[:, np.newaxis],
            "simulated_error_phase": phase
        }



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Interpolation of the SSH from MITGCM interpolated model for WW3
===============================================================
"""
import logging

import numpy as np
import xarray as xr

from .. import data_handler

LOGGER = logging.getLogger(__name__)


class MITGCM_WW3(data_handler.CartesianGridHandler):
    """
    Interpolation of the SSH from MITGCM interpolated for WWW3.
    """
    def __init__(self, path: str):
        loader = MITGCM_WW3.OverriddenNetcdfLoader(
            path,
            date_fmt="%Y%m%d",
            ssh_name="wlv",
            pattern=r"ww3.(?P<date>\d{8})_wlv.nc")
        super().__init__(loader)

    class OverriddenNetcdfLoader(data_handler.NetcdfLoader):
        def load_dataset(self, first_date: np.datetime64,
                         last_date: np.datetime64):
            LOGGER.debug("fetch data for %s, %s", first_date, last_date)
            selected = self.select_netcdf_files(first_date, last_date)
            dataset = xr.open_mfdataset(self.time_series["path"][selected],
                                        concat_dim=self.time_name,
                                        combine="nested",
                                        decode_times=True)

            if self.time_name not in dataset.coords:
                LOGGER.debug(
                    "Time coordinate %s was not found, assigning "
                    "axis with time from file names", self.time_name)
                dataset = dataset.assign_coords(
                    {self.time_name: self.time_series["dates"][selected]})

            LOGGER.debug(
                "Renaming dataset with canonical names lon, lat, time, ssh")
            return dataset.rename({
                self.lon_name: "lon",
                self.lat_name: "lat",
                self.ssh_name: "ssh",
                self.time_name: "time"
            })



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
import datetime
import os
import pathlib
import re
import setuptools
import subprocess


def execute(cmd):
    """Executes a command and returns the lines displayed on the standard
    output"""
    process = subprocess.Popen(cmd,
                               shell=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    return process.stdout.read().decode()


def update_meta(path, version):
    """Updating the version number description in conda/meta.yaml."""
    with open(path, "r") as stream:
        lines = stream.readlines()
    pattern = re.compile(r'{% set version = ".*" %}')

    for idx, line in enumerate(lines):
        match = pattern.search(line)
        if match is not None:
            lines[idx] = '{%% set version = "%s" %%}\n' % version

    with open(path, "w") as stream:
        stream.write("".join(lines))


def update_sphinx_conf(conf, version, year):
    """Update the Sphinx configuration file"""
    with open(conf, "r") as stream:
        lines = stream.readlines()
    pattern = re.compile(r'(\w+)\s+=\s+(.*)')

    for idx, line in enumerate(lines):
        match = pattern.search(line)
        if match is not None:
            if match.group(1) == 'version':
                lines[idx] = "version = %r\n" % version
            elif match.group(1) == 'release':
                lines[idx] = "release = %r\n" % version
            elif match.group(1) == 'copyright':
                lines[idx] = "copyright = '(%s, CNES/JPL)'\n" % year

    with open(conf, "w") as stream:
        stream.write("".join(lines))


def read_version():
    """Returns the software version"""
    module = pathlib.Path('swot_simulator', 'version.py')
    stdout = execute("git describe --tags --dirty --long --always").strip()
    pattern = re.compile(r'([\w\d\.]+)-(\d+)-g([\w\d]+)(?:-(dirty))?')
    match = pattern.search(stdout)

    # If the information is unavailable (execution of this function outside the
    # development environment), file creation is not possible
    if not stdout:
        pattern = re.compile(r'return "(\d+\.\d+\.\d+)"')
        with open(module, "r") as stream:
            for line in stream:
                match = pattern.search(line)
                if match:
                    return match.group(1)
        raise AssertionError("The version module is invalid")

    # No tag already registred
    if match is None:
        pattern = re.compile(r'([\w\d]+)(?:-(dirty))?')
        match = pattern.search(stdout)
        version = "0.0.0"
        sha1 = match.group(1)
    else:
        version = match.group(1)
        commits = int(match.group(2))
        sha1 = match.group(3)
        if commits != 0:
            version += f".dev{commits}"

    stdout = execute("git log  %s -1 --format=\"%%H %%at\"" % sha1)
    stdout = stdout.strip().split()
    date = datetime.datetime.utcfromtimestamp(int(stdout[1]))

    # Conda configuration files are not present in the distribution, but only
    # in the GIT repository of the source code.
    meta = pathlib.Path('conda', 'meta.yaml')
    if meta.exists():
        update_meta(meta, version)

    # Updating the version number description for sphinx
    update_sphinx_conf(pathlib.Path('docs', 'source', 'conf.py'), version,
                       date.year)

    # Finally, write the file containing the version number.
    with open(module, 'w') as handler:
        handler.write('''
# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Get software version information
================================
"""


def release() -> str:
    """Returns the software version number"""
    return "{version}"


def date() -> str:
    """Returns the creation date of this release"""
    return "{date}"
'''.format(version=version, date=date.strftime("%d %B %Y")))
    return version


def main():
    """Main function"""
    here = pathlib.Path(__file__).parent.absolute()
    os.chdir(here)

    with open("README.rst", "r") as fh:
        long_description = fh.read()

    setuptools.setup(
        name="swot_simulator",
        include_package_data=True,
        version=read_version(),
        author="CNES/JPL",
        author_email="fbriol@gmail.com",
        description="Simulate SWOT measurements on sea surface height with "
        "simulated errors",
        long_description=long_description,
        long_description_content_type="text/x-rst",
        url="https://github.com/CNES/swot_simulator",
        packages=setuptools.find_packages(),
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        entry_points='''
    [console_scripts]
    swot_simulator=swot_simulator.launcher:main
    ''',
        python_requires='>=3.6',
        install_requires=[
            "python-dateutil", "distributed", "netCDF4", "numba", "numpy",
            "pyinterp", "scipy", "xarray"
        ])


if __name__ == "__main__":
    main()



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Settings handling
-----------------
"""
from typing import Any, Dict, Iterator, Tuple, Union
import contextlib
import copy
import importlib
import logging
import os
import re
import pathlib
import traceback
import types
import numpy as np
from . import BASIC, DATA, EXPERT, PRODUCT_TYPE, UNSMOOTHED, WIND_WAVE
from . import math
from . import plugins

#: Default working directory
DEFAULT_WORKING_DIRECTORY = os.path.join(os.path.expanduser('~'),
                                         "swot_simulator")

#: Module logger
LOGGER = logging.getLogger(__name__)


def execfile_(filepath: str, _globals: Any) -> None:
    """Executes a Python code defined in a file"""
    with open(filepath, 'rb') as stream:
        source = stream.read()

    code = compile(source, filepath, 'exec')
    exec(code, _globals)


@contextlib.contextmanager
def cd(target_dir: str) -> Iterator[None]:
    """Moves to a directory and returns to the working directory"""
    cwd = os.getcwd()
    try:
        os.chdir(target_dir)
        yield
    finally:
        os.chdir(cwd)


def eval_config_file(filename: str) -> Dict:
    """Evaluate a config file."""
    path = os.path.abspath(filename)
    dirname = os.path.dirname(path)

    namespace = dict(__file__=path)

    with cd(dirname):
        # during executing config file, current dir is changed to ``confdir``.
        try:
            execfile_(filename, namespace)
        except SyntaxError as err:
            raise RuntimeError(
                "There is a syntax error in your configuration file: "
                f"{err}\n") from err
        except SystemExit as err:
            raise RuntimeError(
                "The configuration file (or one of the modules it imports) "
                "called sys.exit()") from err
        except Exception as err:
            raise RuntimeError(
                "There is a programmable error in your configuration "
                f"file:\n\n{traceback.format_exc()}") from err

    return namespace


def error_classes() -> Iterator[str]:
    """Get the list of classes implementing random error generation."""
    module = importlib.import_module(".error", package="swot_simulator")
    for item in dir(module):
        if isinstance(getattr(module, item), type):
            yield item


def error_keywords() -> Iterator[str]:
    """Get the list of the error keywords"""
    for item in error_classes():
        item = item[:1].lower() + item[1:]
        yield re.sub(r'([A-Z])', r'_\1', item).lower()


def _template() -> Tuple[Dict[str, Any], Dict[str, str]]:
    """Get the code representing the default configuration template of the
    simulator."""
    def required_settings():
        """Get the path to the simulator data."""
        result = dict()
        for item in Parameters.REQUIRED:
            result[item] = list()
        for item in os.listdir(DATA):
            for data_type, values in result.items():
                if item.startswith(data_type):
                    values.append(str(DATA.joinpath(item).resolve()))
        return result

    required = required_settings()
    settings = dict()
    for key, (default, _type, _help) in Parameters.CONFIG_VALUES.items():
        if key in required:
            settings[key] = required[key][0]
        elif key == "noise":
            settings[key] = sorted(
                list(set(error_keywords()) - set(("corrected_roll_phase", ))))
        else:
            settings[key] = default
    return required, settings


def _template_to_string(required: Dict[str, str],
                        parameters: Dict[str, Any]) -> str:
    """Get the string representing the default configuration template of the
    simulator."""
    def wrap(s: str) -> Iterator[str]:
        """Cuts the line into several lines of 80 characters maximum"""
        def line(items):
            """Returns the line built"""
            return "# " + " ".join(items)

        words = []
        for item in s.split():
            if sum((len(item) for item in words)) + (len(words) - 1) * 2 > 78:
                yield line(words)
                words.clear()
            words.append(item)
        if words:
            yield line(words)

    result = [
        "import os",
        "import swot_simulator.plugins.ssh",
        "import swot_simulator.plugins.swh",
        "",
        "# Path to the data supplied with the simulator.",
        f"DATA = {str(DATA)!r}",
        "",
    ]
    for key, value in parameters.items():
        result += list(wrap(Parameters.CONFIG_VALUES[key][-1]))
        # The required data is supplied with the simulator. They are all
        # written in the template.
        if key in required:
            # If there are several values for a parameter, only the first one is
            # enabled, the others are disabled.
            enable = True
            for item in required[key]:
                result += [
                    ("" if enable else "# ") + key + " = os.path.join(DATA, " +
                    repr(pathlib.Path(item).name) + ")"
                ]
                enable = False
            result.append("")
        else:
            result += [key + " = " + repr(value), ""]
    # Remove the last carriage return
    result.pop()
    return "\n".join(result)


def template(python: bool = False) -> Union[str, Dict[str, Any]]:
    """Get the template representing the default configuration of the
    simulator

    Args:
        python (bool): True, returns the dictionary that represents the default
            configuration, otherwise returns the Python code of the default
            configuration.

    Returns:
        str, dict: the default configuration of the simulation
    """
    required, dictionary = _template()
    if python:
        return dictionary
    return _template_to_string(required, dictionary)


class NumberOfBeams(int):
    """Handle the number of beams"""
    def __new__(cls, value, *args, **kwargs):
        if value not in [1, 2]:
            raise ValueError("nbeam must be in [1, 2]")
        return super().__new__(cls, value, *args, **kwargs)  # type: ignore


class TimeDelta(int):
    """Handle a time delta in seconds in the configuration file."""
    def __call__(self, value):
        return np.timedelta64(value, "s")


class Seed:
    """Handle the seed for the random state"""
    def __init__(self, seed: int):
        self.seed = seed - 1

    def __call__(self) -> np.random.Generator:
        self.seed += 1
        return np.random.default_rng(seed=self.seed)


class Parameters:
    """
    Simulator parameter management.

    The simulator parameters are defined in a Python file. The expected
    parameters are described in the code below.

    .. literalinclude:: ../settings.py

    Args:
        path (str, optional): Path to the configuration file used to override
            the default settings.
    """
    #: Known parameters.
    CONFIG_VALUES: Dict[str, Tuple[Any, Any, str]] = dict(
        area=(None, [float, 4],
              ("Geographical area to simulate defined by the minimum and "
               "maximum corner point :lon_min, lat_min, lon_max, lat_max. "
               "Default: -180, -90, 180, 90")),
        beam_position=([-20, 20], [float, 2],
                       ("Number of beam used to correct wet troposphere "
                        "signal (1, 2 or 'both')")),
        central_pixel=(False, bool,
                       ("If true, the swath, in the final dataset, will "
                        "contain a center pixel divided in half by the "
                        "reference ground track")),
        complete_product=(False, bool,
                          ("If true, the generated netCDF file will be the "
                           "complete product compliant with SWOT's Product "
                           "Description Document (PDD), otherwise only the "
                           "calculated variables will be written to the "
                           "netCDF file")),
        cycle_duration=(None, float,
                        ("Duration of a cycle in number of fractional days. "
                         "By default, this value is read from the ephemeris "
                         "file)")),
        delta_ac=(2.0, float,
                  ("Distance, in km, between two points across track "
                   "direction")),
        delta_al=(2.0, float,
                  ("Distance, in km, between two points along track "
                   "direction")),
        ephemeris_cols=(None, [int, 3],
                        ("Index of columns to read in the ephemeris file "
                         "containing, respectively, longitude in degrees, "
                         "latitude in degrees and the number of seconds "
                         "elapsed since the start of the orbit. Default: "
                         "[1, 2, 0]")),
        ephemeris=(None, str,
                   ("Ephemeris file to read containing the satellite's "
                    "orbit.")),
        error_spectrum=(None, str,
                        "File containing spectrum of instrument error"),
        corrected_roll_phase_dataset=(None, str,
                                      ("Estimated roll phase dataset. "
                                       "Default: None")),
        half_gap=(10.0, float,
                  ("Distance, in km, between the nadir and the center of the "
                   "first pixel of the swath")),
        half_swath=(60.0, float,
                    ("Distance, in km, between the nadir and the center of "
                     "the last pixel of the swath")),
        height=(None, float, ("Satellite altitude (m). By default, this value "
                              "is read from the ephemeris file.")),
        karin_noise=(None, str,
                     "KaRIN file containing spectrum for several SWH"),
        len_repeat=(20000.0, float, "Repeat length"),
        nadir=(False, bool, "True to generate Nadir products"),
        nbeam=(2, NumberOfBeams,
               ("Number of beam used to correct wet troposphere signal "
                "(1, 2 or 'both')")),
        noise=(None, [str, -1],
               ("This option defined the error to simulate. Allowed values "
                "are \"altimeter,\" \"baseline_dilation,\" "
                "\"corrected_roll_phase,\" \"karin,\" \"orbital,\" "
                "\"roll_phase,\" \"timing,\" and \"wet_troposphere.\" The "
                "calculation of roll errors can be simulated, option "
                "\"roll_phase,\" or interpolated option "
                "\"corrected_roll_phase,\" from the dataset specified by the "
                "value of the option \"roll_phase_dataset.\" Therefore, these "
                "two options are mutually exclusive. In other words, if the "
                "\"roll_phase\" option is present, the "
                "\"corrected_roll_phase\" option must be omitted, and vice "
                "versa.")),
        nseed=(0, int, ("Seed for RandomState. Must be convertible to 32 bit "
                        "unsigned integers")),
        product_type=(EXPERT, str,
                      ("Type of SWOT product to be generated. Possible "
                       f"products are {BASIC!r}, {EXPERT!r}, {UNSMOOTHED!r} "
                       f"and {WIND_WAVE!r}. Default to {EXPERT!r}")),
        requirement_bounds=(None, [float, 2],
                            ("Limits of SWOT swath requirements. Measurements "
                             "outside the span will be set with fill values")),
        shift_lon=(None, float,
                   "Orbit shift in longitude (degrees). Default 0."),
        shift_time=(None, TimeDelta,
                    "Orbit shift in time (seconds). Default 0."),
        sigma=(6.0, float, "Gaussian footprint of sigma (km)"),
        ssh_plugin=(None, plugins.Interface,
                    ("The plug-in handling the SSH interpolation under the "
                     "satellite swath")),
        swh_plugin=(
            None, plugins.Interface,
            "SWH plugin to interpolate model SWH on the SWOT grid. Use only "
            f"{EXPERT!r} or {WIND_WAVE!r} products that have the swh in its "
            "output"),
        swath=(True, bool, "True to generate swath products"),
        swh=(2.0, float, "SWH for the region"),
        working_directory=(DEFAULT_WORKING_DIRECTORY, str,
                           ("The working directory. By default, files are "
                            "generated in the user's root directory")))

    #: Arguments that must be defined by the user.
    REQUIRED = ["ephemeris", "error_spectrum", "karin_noise"]

    def __init__(self, overrides: Dict[str, Any]):
        for required in self.REQUIRED:
            if required not in overrides:
                raise TypeError(f"missing required argument: {required!r}")
        self._init_user_parameters(overrides)

        product_type = getattr(self, "product_type")
        if product_type not in PRODUCT_TYPE:
            raise ValueError(f"Unknown product type: {product_type}")

        if product_type == WIND_WAVE:
            if getattr(self, "ssh_plugin") is not None:
                raise ValueError("The wind/wave product cannot store SSH.")

            if getattr(self, "noise") is not None:
                raise ValueError("The wind/wave product cannot store errors.")

        if product_type == UNSMOOTHED:
            if getattr(self, "central_pixel"):
                raise RuntimeError("The unsmoothed product doesn't support a "
                                   "center pixel.")

        noise = getattr(self, "noise")
        if noise is not None:
            if "corrected_roll_phase" in noise:
                if "roll_phase" in noise:
                    raise TypeError(
                        "option 'corrected_roll_phase' not allowed with option "
                        "'roll_phase'")
                if "corrected_roll_phase_dataset" not in overrides:
                    raise TypeError("missing required argument: "
                                    "'corrected_roll_phase_dataset'")
            noise = [
                "".join(word.capitalize() for word in item.split("_"))
                for item in noise
            ]
            unknowns = set(noise) - set(error_classes())
            if unknowns:
                raise ValueError(
                    f"Unknown error generators: {', '.join(unknowns)}")
        else:
            noise = []
        setattr(self, "noise", noise)
        self._rng = Seed(getattr(self, "nseed"))

    @staticmethod
    def load_default() -> 'Parameters':
        """Load the default configuration

        Returns:
            Parameters: the default settings
        """
        return Parameters(template(python=True))  # type: ignore

    def _convert_overrides(self, name: str, value: Any) -> Any:
        expected_type = self.CONFIG_VALUES[name][1]
        try:
            if isinstance(expected_type, list):
                if not isinstance(value, list):
                    raise ValueError
                length = expected_type[1]
                if length == -1:
                    length = len(value)
                if len(value) != length:
                    raise ValueError
                expected_type = expected_type[0]
                for idx, item in enumerate(value):
                    value[idx] = expected_type(item)
            else:
                mro = expected_type.__mro__[-2]
                if mro == plugins.Interface:
                    value = plugins.Plugin.register(value)
                else:
                    value = expected_type(value)
        except ValueError as exc:
            raise ValueError("invalid value %r for config value %r" %
                             (value, name)) from exc
        return value

    def _init_user_parameters(self, overrides: Dict[str, Any]):
        # To avoid side effects, default values are copied.
        settings = dict((key, copy.copy(value[0]))
                        for key, value in self.CONFIG_VALUES.items())
        for name, value in overrides.items():
            if name in ["__doc__", "__file__", "__builtins__"] or isinstance(
                    value, (types.ModuleType, types.FunctionType)):
                continue
            try:
                if name not in settings:
                    LOGGER.info(
                        'unknown config value %r in override, ignoring', name)
                    continue
                # None values are not processed.
                if value is not None:
                    settings[name] = self._convert_overrides(name, value)
            except ValueError as exc:
                LOGGER.warning("%s", exc)
        self.__dict__.update((key, value) for key, value in settings.items())

    @property
    def box(self) -> math.Box:
        """Get the geographical area of interest."""
        area = self.__dict__["area"]
        if area is None:
            return math.Box()
        return math.Box(math.Point(*area[:2]), math.Point(*area[-2:]))

    def rng(self) -> np.random.Generator:
        """Get a random state generator"""
        return self._rng()



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Logging handlers
----------------
"""
from typing import Awaitable, IO, Optional, Tuple, Union
import logging
import logging.handlers
import pathlib
import pickle
import socket
import struct
import sys
import threading
import tornado.ioloop
import tornado.iostream
import tornado.log
import tornado.tcpserver

#: Synchronize logs for workers
LOCK = threading.RLock()


class LogRecordSocketReceiver(tornado.tcpserver.TCPServer):
    """
    Simple TCP socket-based logging receiver.

    Args:
        name (str, optional): Logger name.
        **kwargs: Base class keyword arguments.
    """
    def __init__(self, name: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self.logname = name

    async def handle_stream(self, stream: tornado.iostream.IOStream,
                            address: Tuple) -> Optional[Awaitable[None]]:
        """Override to handle a new IOStream from an incoming connection.

        Args:
            stream (tornado.iostream.IOStream): Incoming stream.
            address (tuple): IP address and TCP port.
        """
        while True:
            try:
                chunk = await stream.read_bytes(4)
                if len(chunk) < 4:
                    break
                slen = struct.unpack('>L', chunk)[0]
                chunk = await stream.read_bytes(slen)
                while len(chunk) < slen:
                    chunk += await stream.read_bytes(slen - len(chunk))
            except tornado.iostream.StreamClosedError:
                break
            obj = pickle.loads(chunk)
            record = logging.makeLogRecord(obj)
            setattr(record, "ip", address[0])
            self.handle_log_record(record)

    def handle_log_record(self, record: logging.LogRecord) -> None:
        """Handle an incoming logging reccord

        Args:
            record (logging.LogRecord): The event being logged.
        """
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.logname is not None:
            name = self.logname
        else:
            name = record.name
        logger = logging.getLogger(name)

        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)


class LogServer:
    """Handle the log server

    Args:
        hostname (str, optional): Specifies the IP address we want to listen
            to.
        port (int, optional): Specifies the port we want to listen to.
    """
    def __init__(
            self,
            hostname: Optional[str] = None,
            port: int = logging.handlers.DEFAULT_TCP_LOGGING_PORT) -> None:
        hostname = hostname or socket.gethostname()
        self.ip = socket.gethostbyname(hostname)
        self.port = port

        server = LogRecordSocketReceiver()
        server.listen(self.port, self.ip)

        ioloop = tornado.ioloop.IOLoop.current()
        self.thread = threading.Thread(target=ioloop.start)

    def start(self, daemon: bool = False) -> None:
        """Start the server

        Args:
            daemon (bool, optional): Indicates whether this thread is a daemon
                thread (True) or not (False)
        """
        if daemon:
            self.thread.daemon = True
        self.thread.start()

    def __iter__(self):
        yield self.ip
        yield self.port


class LogFormatter(tornado.log.LogFormatter):
    """Inserts the IP address of the worker or scheduler.

    Args:
        *args: Base class arguments.
        **kwarsg: Base class keyword arguments.
    """
    def __init__(self, *args, **kwargs):
        hostname = socket.gethostname()
        self._ip = socket.gethostbyname(hostname)
        super().__init__(*args, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        """Do formatting for a record.

        Args:
            record (logging.Record): information to the event being logged.

        Returns:
            str: The record formatted.
        """
        if "ip" not in record.__dict__:
            record.__dict__['ip'] = self._ip
        return super().format(record)


def _config_logger(stream: Union[IO[str], logging.Handler], level: int,
                   name: str) -> logging.Logger:
    """Configures logbook handler"""
    logger = logging.getLogger(name)
    logger.propagate = True
    formatter = LogFormatter(
        '%(color)s[%(levelname)1.1s - %(ip)s - %(asctime)s - %(module)s]'
        '%(end_color)s %(message)s',
        datefmt='%b %d %H:%M:%S')
    handler = logging.StreamHandler(stream) if not isinstance(
        stream, logging.Handler) else stream
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    if level:
        logger.setLevel(level)
    return logger


def setup(stream: IO[str],
          debug: bool) -> Tuple[logging.Logger, Tuple[str, int, int]]:
    """Setup the logging system

    Args:
        stream (io): Flux used to write in the log.
        debug (bool): True if the log should record debugging traces.

    Returns:
        tuple: The channel logging and log server settings.
    """
    logging_server = LogServer()
    logging_server.start(True)

    if stream is None:
        stream = sys.stdout
    level = logging.DEBUG if debug else logging.INFO
    # Capture dask.distributed
    _config_logger(stream, level=level, name="distributed")
    return _config_logger(
        stream,
        level=level,
        name=pathlib.Path(__file__).absolute().parent.name), (
            logging_server.ip, logging_server.port, level)


def setup_worker_logging(logging_server: Tuple[str, int, int]):
    """Setup the logging server to log worker calculations

    Args:
        logging_server (tuple): Log server connection settings.
    """
    with LOCK:
        for name in [
                "distributed",
                pathlib.Path(__file__).absolute().parent.name
        ]:
            logger = logging.getLogger(name)
            # If this logger is already initialized we do nothing
            if logger.handlers:
                continue
            stream = logging.handlers.SocketHandler(logging_server[0],
                                                    logging_server[1])
            _config_logger(stream, level=logging_server[2], name=name)




# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Get software version information
================================
"""


def release() -> str:
    """Returns the software version number"""
    return "1.2.1"


def date() -> str:
    """Returns the creation date of this release"""
    return "16 September 2021"



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Constants that are defined for SWOT simulator
---------------------------------------------
"""
import pathlib

#
from . import version

#: Module Version
__version__ = version.release()

#: Module release date
__date__ = version.date()

del version

# EARTH CONSTANTS

#: Volumetric mean radius (km)
VOLUMETRIC_MEAN_RADIUS: float = 6371.0

#: Convert degree to km
DEG_2_KM: float = 111.11

#: CELERITY
CELERITY: float = 299800000.0

#: Baseline (m)
BASELINE = 10

#: KA frequency (in Hz)
F_KA = 35750000000.0

#: Basic product
BASIC = "basic"

#: Expert product
EXPERT = "expert"

#: High resolution
UNSMOOTHED = "unsmoothed"

#: Wind-wave product
WIND_WAVE = "wind_wave"

#: Product types
PRODUCT_TYPE = {
    BASIC: "l2b-ssh.xml",
    EXPERT: "l2b-expert.xml",
    UNSMOOTHED: "l2a-highres.xml",
    WIND_WAVE: "l2b-windwave.xml"
}

#: Path to the auxiliary data
DATA = pathlib.Path(__file__).parent / "data"



import os
import tempfile
import dask.distributed
import swot_simulator
import swot_simulator.launcher
import swot_simulator.settings
import swot_simulator.error

ROOT = os.path.dirname(os.path.abspath(__file__))


def test_launcher():
    with tempfile.TemporaryDirectory() as tmpdir:
        parameters = swot_simulator.settings.Parameters.load_default()
        parameters.ephemeris = str(
            swot_simulator.DATA.joinpath("ephemeris_calval_june2015_ell.txt"))
        parameters.nadir = True
        parameters.working_directory = tmpdir
        parameters.noise = [
            'Altimeter', 'BaselineDilation', 'Karin', 'RollPhase', 'Timing'
        ]
        cluster = dask.distributed.LocalCluster()
        client = dask.distributed.Client(cluster)
        client.wait_for_workers(1)
        swot_simulator.launcher.launch(client, parameters, None)
        client.close()
        cluster.close()



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Random signal generation utilities
----------------------------------
"""
from typing import Optional, Tuple
import warnings
import numba as nb
import numpy as np
import scipy.interpolate
import xarray as xr
try:
    import mkl_fft
    IFFT = mkl_fft.ifft
    IFFT2 = mkl_fft.ifft2
except ImportError:
    IFFT = np.fft.ifft
    IFFT2 = np.fft.ifft2


def read_file_instr(file_instr: str, delta_al: float) -> xr.Dataset:
    """Retrieve power spectrum from instrumental noise file provided by
    """
    dataset = xr.load_dataset(file_instr)

    # Set spatial frequency to spatial coordinate
    dataset = dataset.swap_dims(dict(nfreq="spatial_frequency"))

    # Extract spatial frequency relevant to our sampling
    # (Cut frequencies larger than Nyquist frequency)
    cut_min = 1 / (2 * delta_al)

    return dataset.where(dataset.spatial_frequency <= cut_min, drop=True)


def read_file_karin(path: str) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Retrieve power spectrum from instrumental noise file provided by
    """
    with xr.open_dataset(path) as dataset:
        height_sdt = dataset['height_sdt'].data
        cross_track = dataset['cross_track'].data
        swh = dataset['SWH'].data

    return height_sdt, cross_track, swh


@nb.njit(cache=True, nogil=True)
def _interpolate_file_karin(swh_in: np.ndarray, x_ac_in: np.ndarray,
                            height_sdt: np.ndarray, cross_track: np.ndarray,
                            swh: np.ndarray) -> Tuple[float, np.ndarray]:
    """Interpolates the standard deviation of KaRIN instrumental noise as a
    function of SWH and across track distance."""
    warning = 0
    hsdt = np.zeros(swh_in.shape, dtype=np.float64)
    for jx in range(swh_in.shape[1]):
        xacj = abs(x_ac_in[jx])
        indice_ac = np.argmin(np.abs(cross_track - xacj))
        for ix in range(swh_in.shape[0]):
            threshold = swh_in[ix, jx]
            indices = np.argmin(np.abs(swh - threshold))
            if swh[indices] > threshold:
                indices -= 1
            if swh.max() <= threshold:
                hsdt[ix, jx] = height_sdt[-1, indice_ac]
                warning = threshold
            else:
                rswh = threshold - swh[indices]
                hsdt[ix, jx] = height_sdt[indices, indice_ac] * (
                    1 - rswh) + rswh * height_sdt[indices + 1, indice_ac]
    return warning, hsdt


def interpolate_file_karin(swh_in: np.ndarray, x_ac_in: np.ndarray,
                           height_sdt: np.ndarray, cross_track: np.ndarray,
                           swh: np.ndarray) -> np.ndarray:
    """Interpolates the standard deviation of KaRIN instrumental noise as a
    function of SWH and across track distance."""
    if len(swh_in.shape) == 1:
        swh_in = np.expand_dims(swh_in, axis=0)
    warning, hsdt = _interpolate_file_karin(swh_in, x_ac_in, height_sdt,
                                            cross_track, swh)
    if warning:
        warnings.warn(
            f'swh={warning} is greater than the maximum value, '
            f'therefore swh is set to the file maximum '
            'value', RuntimeWarning)
    return hsdt


def gen_psd_1d(fi: np.ndarray,
               psi: np.ndarray,
               rng: np.random.Generator,
               fmin: Optional[float] = None,
               fmax: Optional[float] = None,
               alpha: int = 10,
               lf_extpl: bool = False,
               hf_extpl: bool = False) -> Tuple[np.ndarray, float]:
    """Generate 1d random signal using Fourier coefficient"""
    # Make sure fi, PSi does not contain the zero frequency:
    psi = psi[fi > 0]
    fi = fi[fi > 0]

    # Adjust fmin and fmax to fi bounds if not specified. Values are bounded
    # with respect to the frequencies of the processed spectrum.
    fmin = fmin or fi[0]
    fmax = fmax or fi[-1]

    # Go alpha times further in frequency to avoid interpolation aliasing.
    fmaxr = alpha * fmax

    # Interpolation of the non-zero part of the spectrum
    f = np.arange(fmin, fmaxr + fmin, fmin)
    mask = psi > 0
    ps = np.exp(np.interp(np.log(f), np.log(fi[mask]), np.log(psi[mask])))

    # lf_extpl=True prolongates the PSi as a plateau below min(fi).
    # Otherwise, we consider zeros values. same for hf
    ps[f < fi[0]] = psi[0] if lf_extpl else 0
    ps[f > fi[-1]] = psi[-1] if hf_extpl else 0
    ps[f > fmax] = 0

    # Detect the sections (if any) where PSi==0 and apply it to PS
    mask = np.interp(f, fi, psi)
    ps[mask == 0] = 0

    f_size = f.size
    phase = np.empty((2 * f_size + 1))
    phase[1:(f_size + 1)] = rng.random(f_size) * 2 * np.pi
    phase[0] = 0
    phase[-f_size:] = -phase[1:(f_size + 1)][::-1]

    fft1a = np.concatenate((np.array([0]), 0.5 * ps, 0.5 * ps[::-1]), axis=0)
    fft1a = np.sqrt(fft1a) * np.exp(1j * phase) / np.sqrt(fmin)

    yg = 2 * fmaxr * np.real(IFFT(fft1a))
    return yg, fmaxr


def _gen_signal_1d(fi: np.ndarray,
                   psi: np.ndarray,
                   x: np.ndarray,
                   rng: np.random.Generator,
                   fmin: Optional[float] = None,
                   fmax: Optional[float] = None,
                   alpha: int = 10,
                   lf_extpl: bool = False,
                   hf_extpl: bool = False) -> np.ndarray:
    """Generate 1d random signal using Fourier coefficient"""
    yg, fmaxr = gen_psd_1d(fi, psi, rng, fmin, fmax, alpha, lf_extpl, hf_extpl)
    xg = np.linspace(0, 0.5 / fmaxr * yg.shape[0], yg.shape[0])
    return np.interp(np.mod(x, xg.max()), xg, yg)


def gen_signal_1d(fi: np.ndarray,
                  psi: np.ndarray,
                  x: np.ndarray,
                  rng: np.random.Generator,
                  fmin: Optional[float] = None,
                  fmax: Optional[float] = None,
                  alpha: int = 10,
                  lf_extpl: bool = False,
                  hf_extpl: bool = False) -> np.ndarray:
    """Generate 1d random signal using Fourier coefficient"""
    lf = _gen_signal_1d(fi, psi, x, rng, 1 / 100000000, 1 / 1000000, alpha,
                        True, hf_extpl)
    hf = _gen_signal_1d(fi, psi, x, rng, fmin, fmax, alpha, lf_extpl, hf_extpl)
    return lf + hf


@nb.njit("(float64[:, ::1])"
         "(float64[::1], float64[:, ::1], float64[::1], float64, float64)",
         cache=True,
         nogil=True)
def _calculate_ps2d(f: np.ndarray, f2: np.ndarray, ps1d: np.ndarray,
                    dfx: np.ndarray, dfy: np.ndarray) -> np.ndarray:
    result = np.zeros(f2.shape)
    view = result.ravel()
    dfx_2 = dfx * 0.5
    dfx_y = dfx * dfy
    for idx in range(-1, -f.size - 1, -1):
        item = f[idx]
        mask = (f2 >= (item - dfx_2)) & (f2 < (item + dfx_2))
        amount = np.sum(result[:, idx]) * dfx_y
        miss = ps1d[idx] * dfx - amount
        view[mask.ravel()] = 0 if miss <= 0 else miss * 0.5 / dfx_y
    return result


@nb.njit("(float64[:, ::1])"
         "(float64[:, ::1], float64[::1], float64[::1], float64, float64)",
         cache=True,
         nogil=True)
def _calculate_signal(rectangle, x, y, xgmax, ygmax):
    result = np.zeros((len(y), len(x)))
    xn = (x.max() - x[0]) // xgmax
    yn = (y.max() - y[0]) // ygmax
    dx = x - x[0]
    dy = y - y[0]

    for ix_n in range(int(xn + 1)):
        ix0 = np.where((dx >= (ix_n * xgmax)) & (dx < ((ix_n + 1) * xgmax)))[0]
        for iy_n in range(int(yn + 1)):
            iy0 = np.where((dy >= (iy_n * ygmax))
                           & (dy < ((iy_n + 1) * ygmax)))[0]
            result[iy0[0]:iy0[-1] + 1,
                   ix0[0]:ix0[-1] + 1] = rectangle[:len(iy0), :len(ix0)]
    return result


def gen_ps2d(fi: np.ndarray,
             psi: np.ndarray,
             fminx: float,
             fminy: float,
             fmax: float,
             alpha: int = 10,
             lf_extpl: bool = False,
             hf_extpl: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """TODO(lgaultier)"""
    if fminy < fminx:
        fmin, fminy = fminy, fminx
    else:
        fmin = fminx

    # Go alpha times further in frequency to avoid interpolation aliasing.
    fmaxr = alpha * fmax

    # Make sure fi, PSi does not contain the zero frequency:
    psi = psi[fi > 0]
    fi = fi[fi > 0]

    # Interpolation function for the non-zero part of the spectrum
    f = np.arange(fmin, fmaxr + fmin, fmin)
    ps = np.exp(np.interp(np.log(f), np.log(fi[psi > 0]),
                          np.log(psi[psi > 0])))

    # lf_extpl=True prolongates the PSi as a plateau below min(fi).
    # Otherwise, we consider zeros values. same for hf
    ps[f < fi[0]] = psi[0] if lf_extpl else 0
    ps[f > fi[-1]] = psi[-1] if hf_extpl else 0
    ps[f > fmax] = 0

    # Detect the sections (if any) where PSi==0 and apply it to PS
    psmask = np.interp(f, fi, psi)
    ps[psmask == 0] = 0
    ps1d = ps

    # Build the 2D PSD following the given 1D PSD
    fx = np.concatenate(([0], f))
    fy = np.concatenate(([0], np.arange(fminy, fmaxr + fminy, fminy)))
    fx2, fy2 = np.meshgrid(fx, fy)
    f2 = np.sqrt((fx2**2 + fy2**2))
    dfx = fmin
    dfy = fminy
    ps2d = _calculate_ps2d(f, f2, ps1d, dfx, dfy)
    ps2d[f2 > fmax] = 0
    return ps2d, f


def gen_signal_2d_rectangle(ps2d: np.ndarray,
                            f: np.ndarray,
                            x: np.ndarray,
                            y: np.ndarray,
                            rng: np.random.Generator,
                            fminx: float,
                            fminy: float,
                            fmax: float,
                            alpha: int = 10) -> np.ndarray:
    """TODO(lgaultier)"""
    revert = fminy < fminx
    if revert:
        fmin, fminy = fminy, fminx
        x, y = y, x
    else:
        fmin = fminx

    # Go alpha times further in frequency to avoid interpolation aliasing.
    fmaxr = alpha * fmax

    # Build the 2D PSD following the given 1D PSD
    fx = np.concatenate(([0], f))
    fy = np.concatenate(([0], np.arange(fminy, fmaxr + fminy, fminy)))
    dfx, dfy = fmin, fminy

    phase = rng.random((2 * len(fy) - 1, len(fx))) * 2 * np.pi
    phase[0, 0] = 0.
    phase[-len(fy) + 1:, 0] = -phase[1:len(fy), 0][::-1]

    fft2a = np.concatenate((0.25 * ps2d, 0.25 * ps2d[1:, :][::-1, :]), axis=0)
    fft2a = np.sqrt(fft2a) * np.exp(1j * phase) / np.sqrt((dfx * dfy))
    fft2 = np.zeros((2 * len(fy) - 1, 2 * len(fx) - 1), dtype=complex)
    fft2[:, :len(fx)] = fft2a
    fft2[1:, -len(fx) + 1:] = fft2a[1:, 1:].conj()[::-1, ::-1]
    fft2[0, -len(fx) + 1:] = fft2a[0, 1:].conj()[::-1]

    sg = (4 * fy[-1] * fx[-1]) * np.real(IFFT2(fft2))
    xg = np.linspace(0, 1 / fmin, sg.shape[1])
    yg = np.linspace(0, 1 / fminy, sg.shape[0])
    xgmax = xg.max()
    ygmax = yg.max()

    yl = y - y[0]
    yl = yl[yl < yg.max()]
    xl = x - x[0]
    xl = xl[xl < xg.max()]
    rectangle = np.ascontiguousarray(
        scipy.interpolate.interp2d(xg, yg, sg)(xl, yl))
    signal = _calculate_signal(rectangle, x, y, xgmax, ygmax)

    return signal.transpose() if revert else signal
